from __future__ import annotations


class SetRelation:
    """
    An enum of possible relationships between two sets.
    """

    SUBSET = "subset"

    DISJOINT = "disjoint"

    OVERLAPPING = "overlapping"
